* Program.......: Cw_core.prg
* Authors.......: Joseph D. Booth
* Copyright.....: 1998 Joseph D. Booth, All rights reserved.
* Application...: CLIPWKS++
* Version.......: 1.1
* Compile.......: xpp Cw_core /l /m /w /n
*                 Lc(nRow,nColumn)
*                 Lclose(aWks)
*                 Lr(nTrow,nTcolumn,nBrow,nBcolumn)
*                 Liwks(cFilename|nHandle)
*                 Lversion(cFilename|nHandle)
*                 Lmakerange(cCell_range)
*                 Lcell(cCell)
*                 Lrelative(nCoord1,nCoord2)
*                 Lstring(cString,nLength)
*                 Bread(nBytes,nHandle)
*                 Bwrite(cString,aWks,lFlush)
*
* Special Thanks: Ira Emus -
****************************************************************************
#include "CLIPWKS.CH"
#include "FILEIO.CH"

#define  FLUSH      .T.


STATIC   aErased   := { }

*****************************************************************************

* Function:   Lc()
* Purpose:    To convert row and column coordinates to spreadsheet notation
* Syntax:     Lc( nRow,nColumn [,nSheet] )
* Arguments:  nRow     - Row number
*             nColumn  - Column number
*             nSheet   - Worksheet number, Lotus 3.x only
*
* Notes:      Lc() is used to convert from row and column
*             notation into spreadsheet cell addresses.  It is
*             most often used in loops, as shown in the example below:
*
*             for k=1 to 10
*                goto k              // Goto record 'k'
*                for j=1 to 5
*                   Lput( aWks,fieldget(j),lc(k,j) )
*                next
*             next
*
*             The row may be an integer between 1 and 65536 and
*             the column an integer between 1 and 256.
* Returns:    cCell_address
* See Also:   lr()
********************************
FUNCTION lc( nRow, nColumn, nSheet )

   LOCAL retval := "", temp, remd

   DEFAULT nSheet TO 0

   IF ValType( nRow ) == "N" .AND. ValType( nColumn ) == "N"  // Both parameters numeric?
      IF ( nRow >= 1 .AND. nRow <= MAX_ROWS ) .AND. ;      // Are parameters with
         ( nColumn >= 1 .AND. nColumn <= MAX_COLS )         // range of spreadsheet?

         IF nColumn <= 26                             // Column 26 or less ?
            retval := Chr( nColumn + 64 )

         ELSE
            temp  := Int( nColumn / 26 )                   // Compute first letter
            remd  := nColumn % 26                      // and second letter and
            retval := IF( remd = 0, Chr( temp + 63 ) + "Z", ;
                          Chr( temp + 64 ) + Chr( remd + 64 ) )
         ENDIF

         //retval += ltrim(trim(str(int(nRow),4)))       // Put in row number
         retval += LTrim( Trim( Str( Int( nRow ), 5 ) ) )       // Put in row number

         IF PCount() = 3

            IF nSheet > 0 .AND. nSheet <= 26              // Lotus 3, spreadsheet
               retval := Chr( nSheet + 64 ) + ":" + retval        // number
            ELSE
               temp   := Int( nSheet / 26 )                   // Compute first letter
               remd   := nSheet % 26                      // and second letter and
               retval := IF( remd = 0, Chr( temp + 63 ) + "Z", ;
                             Chr( temp + 64 ) + Chr( remd + 64 ) ) + ":" + retval
            ENDIF
         ENDIF

      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lc *
*****************************************************************************



* Function:   Lclose()
* Purpose:    To write the EOF op code and close the file handle
* Syntax:     Lclose( aSpreadsheet )
* Arguments:  aSpreadsheet     - Spreadsheet handle array
* Notes:      After any spreadsheet is opened, either through
*             Lcreate() or Lread(), you must close it to release
*             its buffer and its handle.  If you do not close a
*             newly created spreadsheet, there is a good chance
*             that some data will not be written out to the
*             disk. Be sure to close all spreadsheets after you
*             are done with them.
* Returns:    lSuccess
*
*********************
FUNCTION lclose( aWks )

   LOCAL retval := ERR_INVALID_PARAMETERS

   IF ValType( aWks ) == "A" .AND. ! Empty( aWks )       // Valid parameter ?

      IF Empty( aErased )
         IF ValType( WKS_TYPE ) == "C"                // If new file, then
            IF SubStr( WKS_VERSION, 1, 1 ) == "E"
               Bwrite( E2_EOF_OPCODE, aWks, FLUSH )     // add an EOF opcode
            ELSE
               Bwrite( EOF_OPCODE, aWks, FLUSH )        // add an EOF opcode
            ENDIF
         ENDIF
      ELSE
         // Re-write the spreadsheet
      ENDIF

      FClose( WKS_HANDLE )                            // Close the worksheet
      retval := IF( FError() = 0, ALL_OK, ERR_IO_ERROR )  // Update return status
      aWks   := { }                                  // Clear out the array
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lclose *
*****************************************************************************



* Function:   Lext()
* Purpose:    To determine the file extension for the spreadsheet version
* Syntax:     Lext( cVersion )
* Arguments:  cVersion  - L1  Lotus 1.x
*                       - L2  Lotus 2.x
*                       - L3  Lotus 3.x
*                       - QU  Quattro
*                       - QP  Quattro Pro
*                       - E2  Excel
*                       - E3  Excel for Windows
* Returns:    cExtension
**************************
FUNCTION lext( cVersion )
   LOCAL cExt := ".WK1"
   DO CASE
   CASE cVersion == "QP"
      cExt := ".WQ1"
   CASE cVersion == "QU"
      cExt := ".WKQ"
   CASE cVersion == "L1"
      cExt := ".WKS"
   CASE cVersion == "L3"
      cExt := ".WK3"
   CASE cVersion == "L4"
      cExt := ".WK4"
   CASE cVersion == "E2" .OR. cVersion == "E3" .OR. cVersion == "E4"
      cExt := ".XLS"
   OTHERWISE
      cExt := ".WK1"
   ENDCASE
   RETURN cExt
*****************************************************************************
* End of Lext *
*****************************************************************************


* Function:   Lcformat()
* Purpose:    To convert a cell format into an integer byte value
* Syntax:     Lcformat( cString )
* Arguments:  cString  - D1  yy-mmm-dd
*                      - D2  dd-mmm
*                      - D3  mmm-yy
*                      - Fx  Fixed, x decimal places
*                      - Sx  Sci Notation, x decimal places
*                      - Cx  Currency, x decimal places
*                      - Px  Percent, x decimal places
*                      - ,x  Comma, x decimal places
* Returns:    nValue
*********************************

FUNCTION lcformat( cString, aWks, lProtect )
   LOCAL nValue   := 0
   LOCAL cType    := SubStr( cString, 1, 1 )
   LOCAL nPlaces  := Val( SubStr( cString, 2 ) )

   IF WKS_VERSION  == "E3" .OR. WKS_VERSION == "E4"
      DO CASE
      CASE cString == "D2"
         nValue := 20
      CASE cString == "D3"
         nValue := 21
      CASE cString == "P0"
         nValue := 22
      CASE SubStr( cString, 1, 1 ) == "P"
         nValue := 23
      CASE cString == ",0"
         nValue := 24
      CASE SubStr( cString, 1, 1 ) == ","
         nValue := 25
      CASE cString == "C0"
         nValue := 26
      CASE SubStr( cString, 1, 1 ) == "C"
         nValue := 27
      CASE SubStr( cString, 1, 1 ) == "S"
         nValue = 28
      CASE cString == "F0"
         nValue := 29
      CASE SubStr( cString, 1, 1 ) == "F"
         nValue := 30
      ENDCASE
   ELSEIF WKS_VERSION == "E2"
      DO CASE
      CASE cString == "D1"
         nValue := 13
      CASE cString == "D2"
         nValue := 15
      CASE cString == "D3"
         nValue := 16
      CASE cString == "P0"
         nValue := 10
      CASE SubStr( cString, 1, 1 ) == "P"
         nValue := 11
      CASE cString == ",0"
         nValue := 4
      CASE SubStr( cString, 1, 1 ) == ","
         nValue := 5
      CASE cString == "C0"
         nValue := 6
      CASE SubStr( cString, 1, 1 ) == "C"
         nValue := 7
      CASE SubStr( cString, 1, 1 ) == "S"
         nValue = 12
      CASE cString == "F0"
         nValue := 2
      CASE SubStr( cString, 1, 1 ) == "F"
         nValue := 3
      ENDCASE
   ELSE
      IF lProtect == NIL
         IF ( WKS_PROTECTION == "P" )              // If protection has been enabled
            nValue += 128
         ENDIF
      ELSE
         IF lProtect
            nValue += 128
         ELSE
            IF nValue > 127
               nValue -= 128
            ENDIF
         ENDIF
      ENDIF

      IF cType == "D"                            // A date format cell
         nValue += ( 113 + nPlaces )
      ELSE
         nValue += At( cType, "SCP," ) * 16 + nPlaces   // Other cell formats, defaults
      ENDIF                                      // to FIXED
   ENDIF
   RETURN nValue
*****************************************************************************
* End of LcFormat *
*****************************************************************************



* Function:   Lr()
* Purpose:    To convert row and column coordinates to spreadsheet range
* Syntax:     Lr( nTopRow,nTopColumn,nBottomRow,nBottomColumn[,nSheet] )
* Arguments:  nTopRow       - Row number
*             nTopColumn    - Column number
*             nBottomRow    - Bottom row number
*             nBottomColumn - Bottom column number
*             nSheet        - Spreadsheet number, Lotus 3.x only
* Notes:      Lr() takes four coordinates and converts them into
*             a cell address range.  It is used to allow the
*             program to work in terms of numeric row and
*             columns and let CLIPWKS handle the translation
*             into spreadsheet syntax.
* Returns:    cRange_address
* See Also:   Lc()
*************************************************
FUNCTION lr( nTrow, nTcolumn, nBrow, nBcolumn, nSheet )

   LOCAL part_1, part_2
   DEFAULT nSheet TO 0
   IF PCount() = 5
      part_1 := lc( nTrow, nTcolumn, nSheet )              // Pass coordinates to
      part_2 := lc( nBrow, nBcolumn, nSheet )              // lc() function call
   ELSE
      part_1 := lc( nTrow, nTcolumn )                     // Pass coordinates to
      part_2 := lc( nBrow, nBcolumn )                     // lc() function call
   ENDIF

   IF ! Empty( part_1 ) .AND. ! Empty( part_2 )         // If both are valid
      RETURN part_1 + ".." + part_2                 // concatenate with ..
   ENDIF
   RETURN ""
*****************************************************************************
* End of Lr *
*****************************************************************************

* Function:   Liswks()
* Purpose:    To test if a file is a valid spreadsheet
* Syntax:     Liswks( cFile|nHandle )
* Arguments:  cFile         - File name
*             nHandle       - Handle file is opened on
* Notes:      Liswks() tests the beginning and ending bytes of
*             the file to see if the file appears to be a valid
*             spreadsheet file.  It is useful to confirm the
*             validity of a file if your application allows a
*             user to type in a spreadsheet name.
* Returns:    lSuccess
* See Also:   Lversion()
**************************
FUNCTION liswks( xFilename )
   RETURN ! Empty( lVersion( xFilename ) )
*****************************************************************************
* End of Liswks *
*****************************************************************************

* Function:   Lversion()
* Purpose:    To determine the version of the spreadsheet
* Syntax:     Lversion( cFile|nHandle )
* Arguments:  cFile         - File name
*             nHandle       - Handle file is opened on
* Notes:      If the file|handle appears to be a valid
*             spreadsheet, this function will return a two
*             character code indicating the version of
*             spreadsheet it appears to be.
*
* Returns:    cVersion      - L1  Lotus 1.x
*                           - L2  Lotus 2.x
*                           - L3  Lotus 3.x
*                           - QU  Quattro
*                           - QP  Quattro Pro
*                           - E2  Excel
*                           - E3  Excel for Windows
* See Also:   Liswks()
****************************
FUNCTION lversion( xFilename )

   LOCAL retval := "", wkshandle := 0, nWhere, opcode, length, close_it := .T., ver

   IF ValType( xFilename ) == "C"

      nWhere    := At( ".", xFilename )
      xFilename := IF( nWhere = 0, Trim( xFilename ) + "." + DEFAULT_EXTENSION, xFilename )
      IF File( xFilename )
         wkshandle := FOpen( xFilename, FO_READ )  // Open file read-only
      ENDIF

   ELSE

      wkshandle := xFilename                      // Numeric parameter is a file
      close_it  := .F.                            // handle which shouldn't be
   ENDIF                                          // closed when done.

   IF wkshandle >= 0

      FSeek( wkshandle, 0, FS_SET )                   // Go to top of file
      Bread(, wkshandle )                           // Initialize the buffer
      opcode := Bin2W( bread( 2, wkshandle ) )           // Read operation code
      length := Bin2W( bread( 2, wkshandle ) )           // and data length

      ver   := Bin2W( bread( 2, wkshandle ) )
      DO CASE
      CASE ver == 1028                             // Lotus 1A
         retval := "L1"
      CASE ver == 1029                             // Symphony 1.0
         retval := "SY"
      CASE ver == 1030                             // Lotus 2.0a-2.2
         retval := "L2"
      CASE ver == 4096                             // Lotus 3.0
         retval := "L3"
      CASE ver == 4098                             // Lotus 4.0
         retval := "L4"
      CASE ver == 1536                             // Lotus 123/J
         retval := "LJ"
      CASE ver == 2571                             // Quattro
         retval := "QU"
      CASE ver == 20768                            // Quattro Pro
         retval := "QP"
      CASE opcode = 2057 .AND. ver = 1280          // Excel 5.x
         retval := "E5"
      CASE ver == 0 .AND. opcode = 1033            // Excel 4.x
         retval := "E4"
      CASE ver == 0 .AND. opcode =  521            // Excel 3.x
         retval := "E3"
      CASE ver == 2
         retval := "E2"
      ENDCASE

      Bread(, - 1 )

      IF close_it
         FClose( wkshandle )
      ENDIF
   ENDIF
   RETURN retval
*****************************************************************************
* End of Lversion *
*****************************************************************************

* Function:   Lmakerange()
* Purpose:    To convert a cell range to it internal format
* Syntax:     Lmakerange( cRange,cVersion )
* Arguments:  cRange      - range of cells ( address .. address )
*             cVersion    - type of spreadsheet
* Notes:      Lmakerange() is used internally to convert
*             spreadsheet range syntax into its internal format
*             for use by other programs.  It should never be
*             called directly by your application.  It is
*             included here in case you need to modify CLIPWKS.
* Returns:    cInternal   - Internal format of the range
* Class:      Internal
************************************
FUNCTION lmakerange( cell_range, cVer )

   LOCAL _starting := "", _ending := "", k, which := "S", temp, retval := ""
   LOCAL cSheet1 := "", cSheet2 := ""
   //LOCAL cMiddle := IF( SubStr( cVer, 1, 1 ) == "E", ":", ".." )
   cell_range := StrTran( cell_range, "$" )
   IF cVer == "L3" .OR. cVer == "L4"
      k := At( ":", cell_range )
      IF k > 0
         cSheet1 := SubStr( cell_range, 1, k )
         cell_range := SubStr( cell_range, k + 1 )
         k := At( ":", cell_range )
         IF k > 0
            cSheet2 := SubStr( cell_range, k - 1, 2 )
            cell_range := SubStr( cell_range, 1, k - 2 ) + SubStr( cell_range, k + 1 )
         ENDIF
      ENDIF
   ENDIF

   FOR k = 1 TO Len( cell_range )
      temp := Upper( SubStr( cell_range, k, 1 ) )

      IF ( temp >= "A" .AND. temp <= "Z" ) .OR. temp $ "0123456789"

         IF which == "S"
            _starting += temp
         ELSE
            _ending   += temp
         ENDIF

      ELSE

         which := "E"                  // .. found, so we start the end part of
         // the range
      ENDIF
   NEXT

   IF ! Empty( _starting ) .AND. ! Empty( _ending )
      retval := Lcell( cSheet1 + _starting, cVer ) + Lcell( cSheet2 + _ending, cVer )
   ENDIF

   RETURN retval
*****************************************************************************
* End of Lmakerange *
*****************************************************************************

* Function:   Lcell()
* Purpose:    To convert a cell address to it internal format
* Syntax:     Lcell( cAddress,cVersion )
* Arguments:  cAddress    - Address of cell
*             cVersion    - type of spreadsheet
* Notes:      Lcell() is used to convert a cell in spreadsheet
*             notation to its internal representation in the
*             file. For the most part, you should never need to
*             call Lcell() directly.
* Returns:    cInternal   - Internal format of the range
* Class:      Internal
**************************
FUNCTION lcell( _cell, cVer )

   LOCAL _cellstring := "", _col := "", _row := "", _bit, _coln, _rown
   LOCAL nSheet := 0, x

   _cell := Upper( _cell )

   IF ( x := At( ":", _cell ) ) > 0
      IF x = 2
         nSheet := Asc( SubStr( _cell, 1 ) ) - 65
      ELSE
         nSheet := ( Asc( SubStr( _cell, 1 ) ) - 64 ) * 26 + ;
                   ( Asc( SubStr( _cell, 2, 1 ) ) - 65 )
      ENDIF
      _cell  := SubStr( _cell, x + 1 )
   ENDIF

   WHILE ! Empty( _cell )

      _bit := SubStr( _cell, 1, 1 )

      IF _bit >= "A" .AND. _bit <= "Z"
         _col += _bit
      ELSEIF _bit >= "0" .AND. _bit <= "9"
         _row += _bit
      ENDIF

      _cell := SubStr( _cell, 2 )

   ENDDO
   _rown     := Val( _row ) - 1
   _coln     := IF( Len( _col ) = 1, Asc( _col ) - 65, ( Asc( _col ) - 64 ) * 26 + ( Asc( SubStr( _col, 2, 1 ) ) - 65 ) )

   IF _rown <= MAX_ROWS - 1 .AND. _coln <= MAX_COLS - 1
      IF SubStr( cVer, 1, 1 ) == "E"
         _cellstring := I2Bin( _rown ) + I2Bin( _coln )
      ELSE
         _cellstring := IF( cVer == "L3" .OR. cVer == "L4", ;
                            I2Bin( _rown ) + Chr( nSheet ) + Chr( _coln ), ;
                            I2Bin( _coln ) + I2Bin( _rown )  )
      ENDIF

   ENDIF

   RETURN _cellstring
*****************************************************************************
* End of Lcell *
*****************************************************************************


* Function:   Lrelative()
* Purpose:    To convert cell coordinates to relative values
* Syntax:     Lrelative( nRefer,nCurrent )
* Arguments:  nRefer       - Row|column referred to
*             nCurrent     - Current row|column
* Notes:      Lrelative() is used to compute relative
*             coordinates between two cell address.
*             Spreadsheets store formula cell references are
*             relative positions from the current cell.  This
*             allows copying of formulas to automatically adjust
*             cell reference.
* Returns:    nRelative
* Class:      Internal
**********************************************
FUNCTION lrelative( _p1, _p2 )
   LOCAL diff := ( _p1 - _p2 )
   RETURN IF( diff = 0, 32768, IF( diff < 0, 49152 + diff, 32768 + diff ) )
*****************************************************************************
* End of Lrelative *
*****************************************************************************



* Function:   Lstring()
* Purpose:    To convert a character string to spreadsheet format
* Syntax:     Lstring( cString, nSize )
* Arguments:  cString      - String to write to spreadsheet
*             nSize        - Size string should be
* Notes:      Spreadsheet character strings are stored with a
*             trailing NULL.  This function takes a string and
*             places a CHR(0) on the end so it can be properly
*             written to the spreadsheet.
* Returns:    cString
* Class:      Internal
*******************************
FUNCTION lstring( cString, nSize )
   RETURN SubStr( cString + Replicate( Chr( 0 ), nSize ), 1, nSize - 1 ) + Chr( 0 )
*****************************************************************************
* End of Lstring *
*****************************************************************************



* Function:   Bread()
* Purpose:    To read the requested number of bytes from a data file buffer
* Syntax:     Bread( nBytes, nHandle )
* Arguments:  nBytes       - number of bytes to be read
*             nHandle      - Handle the file is opened on
* Notes:      Bread() is used to perform buffered reading of a
*             spreadsheet file. The 2048 bytes in the buffer are
*             used first to satisfy each read request, allowing
*             the functions to operate more quickly.  Bread()
*             should never be called directly by your
*             application.
* Returns:    cString      - data buffer requested
* Class:      Internal
******************************
FUNCTION bread( nBytes, nHandle )               // Thanks - Ira Emus

   STATIC buffer, pointer, offset
   LOCAL temp, retval

   IF nBytes == NIL

      IF nHandle = - 1                           // Release buffer
         buffer := ""
         RETURN offset + pointer                  // Where I am in the file
      ELSE
         pointer := 1                           // Initialize buffer
         offset  := FSeek( nHandle, 0, 1 )          // Save the file offset
         buffer  := Space( BUFFSIZE )
         FRead( nHandle, @buffer, BUFFSIZE )
      ENDIF
      RETURN NIL

   ENDIF

   IF nBytes + pointer > BUFFSIZE + 1
      temp    := SubStr( buffer, pointer )
      nBytes  := pointer + nBytes - ( BUFFSIZE + 1 )
      pointer := 1
      offset  := FSeek( nHandle, 0, 1 )             // Save the file offset
      FRead( nHandle, @buffer, BUFFSIZE )
      retval  := temp + SubStr( buffer, 1, nBytes )
   ELSE
      retval  := SubStr( buffer, pointer, nBytes )
   ENDIF
   pointer += nBytes
   RETURN retval
*****************************************************************************
* End of Bread *
*****************************************************************************


* Function:   Bwrite()
* Purpose:    To write the requested string to a spreadsheet buffer
* Syntax:     Bwrite( cString,aSpreadsheet,lFlush )
* Arguments:  cString      - String to write to the buffer
*             aSpreadsheet - Array returned when spreadsheet was opened
*             lFlush       - Force buffer to be written to disk?
* Notes:      Bwrite() is used to write to the buffer rather
*             than a disk file.  It allows for better
*             performance when creating spreadsheets.  You
*             should never call Bwrite() directly in your
*             application code.
* Returns:    NIL
* Class:      Internal
**************************
FUNCTION bwrite( cString, aWks, lFlush )

   DEFAULT lFlush TO .F.
   WKS_BUFFER += cString                      // Add string to write buffer

   IF Len( WKS_BUFFER ) > BUFFSIZE .OR. lFlush  // If past buffer size, write out

      Ldirty( aWks )                            // Check for parameters to be written
      FWrite( WKS_HANDLE, WKS_BUFFER )           // Write the actual buffer
      WKS_BUFFER := ""                        // and clear it out

   ENDIF
   RETURN NIL
*****************************************************************************
* End of Bwrite *
*****************************************************************************



* Function:   Treal()
* Purpose:    To convert a real number to/from Lotus 3.x format
* Syntax:     treal( nValue | cValue )
* Arguments:  nValue  - real number to convert to IEEE real long 10 byte
*             cValue  - IEEE real long 10 byte string to convert to numeric
* Notes:      Treal() is used to read/write Lotus 3.x 10 byte IEEE real
*             numbers.
* Returns:    cResult | nResult
* Class:      Internal
************************
FUNCTION treal( xValue, lSpecial )

   //LOCAL cTmp := "",k,nDiv, tmp, nDec := 0
   //LOCAL nSign, nExp
   //LOCAL retval
   IF ValType( xValue ) == "N"
      RETURN l_dtot( xValue )          // Thanks Rob Hannah
   ENDIF
   IF( SubStr( xValue, 9, 2 ) == Chr( 255 ) + Chr( 255 ) )
   lSpecial := Asc( SubStr( xValue, 8, 1 ) ) == 224
   RETURN 0
ENDIF
RETURN l_ttod( xValue )               // Thanks Rob
*****************************************************************************
* End of Treal *
*****************************************************************************


* Function:   Shortnum()
* Purpose:    To convert a Lotus 3.x integer to/from a Clipper numeric
* Syntax:     Shortnum( nValue|cValue )
* Arguments:  nValue   - Integer to convert to short number
*             cValue   - Character string to convert to integer
* Notes:      Shortnum() is used to read/write Lotus 3.x short number
*             cells.
* Returns:    cResult | nResult
* Class:      Internal
****************************
FUNCTION shortnum( xValue )

   LOCAL retval := ""

   IF ValType( xValue ) == "N"
      retval := IF( xValue >= 0, I2Bin( xValue * 2 ), I2Bin( 65536 - ( Abs( xValue ) * 2 ) ) )
   ELSEIF ValType( xValue ) == "C"
      retval := l_snum( xValue )         // Thanks Rob Hannah
   ENDIF

   RETURN retval
*****************************************************************************
* End of Shortnum() *
*****************************************************************************



* Function:   Real2iee()
* Purpose:    To convert a real number from Lotus/Quattro format
* Syntax:     Real2iee( cValue )
* Arguments:  cValue  - IEEE real long 8 byte string to convert to numeric
* Returns:    nResult
* Class:      Internal
***************************
FUNCTION real2iee( cValue, lSpecial )

   LOCAL x := Asc( SubStr( cValue, 8, 1 ) ), y := Asc( SubStr( cValue, 7, 1 ) )
   IF y == 240 .AND. ( x == 127 .OR. x == 255 )     // Check for ERR/NA cell
      lSpecial := SubStr( cValue, 1, 6 ) <> Replicate( Chr( 0 ), 6 )
      RETURN 0
   ENDIF
   RETURN realiee( cValue )               // Thanks, Mike Taylor
*****************************************************************************
* End of Real2iee() *
*****************************************************************************
