* Program.....: Export.prg
* Authors.....: Joseph D. Booth
* Copyright...: 1998 Joseph D. Booth, All rights reserved.
* Application.: CLIPWKS++
* Version.....: 1.1
* Compile.....: xpp export -l -m -n -w
*
****************************************************************************
#include "CLIPWKS.CH"
//
// Taken from DBSTRUCT.CH
//
#define  DBS_NAME      1  // Field name
#define  DBS_TYPE      2  // Field type
#define  DBS_LEN       3  // Field length
#define  DBS_DEC       4  // Number of decimal places


* Function:   Lexport()
* Purpose:    To export a .DBF file to a spreadsheet
* Syntax:     Lexport( cSpreadsheet,aFld_exp,aHdg_exp,cRecords,cVersion )
* Arguments:  cSpreadsheet - name of the spreadsheet to create
*             aFld_exp     - array of fields to export
*             aHdg_exp     - array of heading expressions
*             cRecords     - which records to write to spreadsheet
*             cVersion     - type of spreadsheet to produce
* Notes:      Lexport() is a high-level function use to transfer
*             records from a .DBF file into a spreadsheet.  A
*             great degree of flexibility is provided, including
*             control of the fields and headings, and the
*             selection of records to write to the spreadsheet.
*
*             cWksfile is the name of the spreadsheet to create
*             and it defaults to the work area name with a .WK1
*             extension.
*
*             aFields is either an array or a semi-colon
*             delimited list of fields/functions to place in
*             each column of the spreadsheet.
*
*             aHeaders is a corresponding array or semi-colon
*             delimited list of header strings.  Each element
*             contains the header for the parallel entry in the
*             aFields array.  If aHeaders is a logical .F., no
*             headers will be written to the spreadsheet.
*
*             cRecs is a character string indicating which
*             records to transfer into the spreadsheet.
*
*             cVersion may be
*
*             L2  - Lotus 2.x
*             L3  - Lotus 3.x
*             QU  - Quattro
*             QP  - Quattro Pro
* Returns:    nStatus
**********************************************
FUNCTION lexport( cFilename, fldlist, hdglist, which_recs, cVersion )

   LOCAL temp, k, fc, headers := .T., x, aWks, rc, sk_fact, fw
   LOCAL _seekit, ret_val := 0, l3dates := { }, _cond, bBlock

   DEFAULT cVersion TO "L2"             // Default to Lotus 2.x spreadsheets

   IF "" = Alias()                      // The work area is empty
      RETURN - 1                         // so return an error code
   ENDIF
***********************************
* Validate file name *
***********************************
   IF ValType( cFilename ) <> "C"
      cFilename := Trim( Alias() )
   ENDIF
/********************************************************
* Convert semi-colon delimited string to an array *
********************************************************/
   IF ValType( fldlist ) == "C"                // if a semi-colon delimited string
      IF ! Empty( fldlist )                   // is passed, convert it to an array
         fldlist := to_arr( fldlist )
      ENDIF
   ENDIF
***************************************************
* If not passed, build default fldlist file *
***************************************************
   IF ValType( fldlist ) <> "A"
      fldlist := { }
      temp    := DBStruct()
      FOR k   := 1 TO Len( temp )
         IF temp[ k, DBS_TYPE ] = "D"
            temp[ k, DBS_LEN ] ++
            AAdd( l3dates, k )
         ENDIF
         AAdd( fldlist, Trim( temp[ k, DBS_NAME ] ) + "|" + ;
               Str( temp[ k, DBS_LEN ], 3 ) + "." + ;
               LTrim( Trim( Str( temp[ k, DBS_DEC ], 2 ) ) ) )
      NEXT
   ENDIF
   fc := Len( fldlist )
*********************************************************
* Convert semi-colon delimited string to an array *
*********************************************************

   IF ValType( hdglist ) == "C"                // if a semi-colon delimited string
      IF ! Empty( hdglist )                   // is passed, convert it to an array
         hdglist := to_arr( hdglist )
      ENDIF
   ELSEIF ValType( hdglist ) == "L"            // Logical .F. disables headers
      headers := hdglist
   ENDIF

   IF headers .AND. ValType( hdglist ) <> "A"  // Headers needed, but not passed
      hdglist := { }
      FOR k := 1 TO fc
         x := At( "|", fldlist[ k ] )
         AAdd( hdglist, IF( Empty( x ), fldlist[ k ], SubStr( fldlist[ k ], 1, x - 1 ) ) )
      NEXT
   ENDIF

   IF ValType( which_recs ) <> "A"
      which_recs := IF( ValType( which_recs ) == "C", which_recs, "ALL" )
   ENDIF

****************************************************************************
* Attempt to create new file *
****************************************************************************

   aWks       := Lcreate( cFilename, cVersion ) // Create a spreadsheet file

   IF ! Empty( aWks )                             // If a valid handle returned
   *
      DO CASE                                  // Try to determine number of
      CASE ValType( which_recs ) == "A"            // records to write to lotus
         rc := Len( which_recs )

      CASE which_recs == "ALL"
         rc := RecCount()                      // Number of records

      CASE SubStr( which_recs, 1, 5 ) = "EVERY"

         sk_fact  := Val( SubStr( which_recs, 7 ) ) // Extract number of records
         sk_fact  := IF( sk_fact = 0, 2, sk_fact )   // to skip.  Default to 2
         rc       := IF( headers, 2, 0 ) + ( RecCount() / sk_fact ) + 1

      CASE SubStr( which_recs, 1, 3 ) = "TAG"
         rc := Len( Trim( which_recs ) ) - Len( StrTran( Trim( which_recs ), "", "" ) ) + 1

      OTHERWISE
         rc       := IF( headers, 2, 0 ) + RecCount()

      ENDCASE
   *
      Lrange( aWks, rc + IF( headers, 2, 0 ), fc )     // Write record range to file
   *
      FOR k := 1 TO fc
         fw := Lgetwide( fldlist[ k ] )
         Lwidth( aWks, k, fw + 1 )                   // Set width of column
      NEXT
   *
      IF cVersion $ "L3|L4"                    // Format Lotus 3.x + dates
         FOR k := 1 TO Len( l3dates )
            Lformat( aWks, lr( IF( headers, 3, 1 ), l3dates[ k ], IF( headers, rc + 2, rc ), l3dates[ k ] ), "D1" )
         NEXT
      ENDIF
   *
      IF headers                              // Check for headers
         FOR k := 1 TO fc                       // Write headings into file
            Lput( aWks, hdglist[ k ], lc( 1, k ) )     // Field name
            Lput( aWks, "\=", lc( 2, k ) )           // Underline the field name
         NEXT
      ENDIF
*********************************************************
* Transfer each record into the lotus spreadsheet *
*********************************************************
      k := IF( headers, 0, - 2 )
      DO CASE
      CASE ValType( which_recs ) == "A"
         FOR x := 1 TO rc
            goto which_recs[ x ]
            Lputout( ++ k, fldlist, aWks )
         NEXT
      CASE which_recs == "ALL"
         go top
         WHILE ! Eof()
            Lputout( ++ k, fldlist, aWks )
            skip + 1
         ENDDO
      CASE SubStr( which_recs, 1, 5 ) == "EVERY"
         go top
         WHILE ! Eof()
            Lputout( ++ k, fldlist, aWks )
            skip sk_fact
            rc ++
         ENDDO
      CASE SubStr( which_recs, 1, 3 ) == "TAG"
         FOR x = 1 TO Len( which_recs ) + 4
            goto x
            IF SubStr( which_recs, x + 4, 1 ) = ""
               Lputout( ++ k, fldlist, aWks )
            ENDIF
         NEXT
      CASE SubStr( which_recs, 1, 3 ) == "FOR"
         _cond := SubStr( which_recs, 5 )
         bBlock := &( "{ || " + AllTrim( _cond ) + " }" )
         go top
         IF Eval( bBlock ) <> "L"
            ret_val := - 3
         ELSE
            WHILE ! Eof()
               IF Eval( bBlock )
                  Lputout( ++ k, fldlist, aWks )
                  rc ++
               ENDIF
               skip + 1
            ENDDO
         ENDIF

      CASE SubStr( which_recs, 1, 5 ) == "WHILE"
         which_recs := SubStr( which_recs, 7 )
         x          := At( ",", which_recs )
         _seekit    := SubStr( which_recs, 1, x - 1 )
         _cond      := SubStr( which_recs, x + 1 )
         bBlock     := &( "{ || " + AllTrim( _cond ) + " }" )
         go top
         seek _seekit
         IF Eof()
            ret_val := - 4
         ELSEIF Eval( bBlock ) <> "L"
            ret_val := - 3
         ELSE
            WHILE ! Eof() .AND. Eval( bBlock )
               Lputout( ++ k, fldlist, aWks )
               rc ++
               skip + 1
            ENDDO
         ENDIF
      ENDCASE
*****************************
* Close the worksheet *
*****************************
      Lclose( aWks )
   ELSE
      ret_val := - 2
   ENDIF
   RETURN ret_val
* *************************




* Function:    Lputout()
* Purpose:     To write a row of data into a spreadsheet
* Syntax:      Lputout( nRow,aField_list,aSpreadsheet )
* Arguments:   nRow         - Row number to put data into
*              aField_list  - Array of field expressions to write
*              aSpreadsheet - array of spreadsheet parameters
* Notes:       Lputout() is a STATIC function used by Lexport()
*              to write data into a cell.  It should never be
*              called directly by your application.  It is used
*              to write all the fields/functions from the current
*              record into the spreadsheet.
* Returns:     1
* Class:       Internal
**********************************************
STATIC FUNCTION lputout( nRow, flist, aWks )
   LOCAL j, jj := 0, x, tmp, bBlock

   FOR j = 1 TO Len( flist )
      x       := At( "|", flist[ j ] )
      tmp     := IF( x > 0, SubStr( flist[ j ], 1, x - 1 ), flist[ j ] )
      bBlock  := &( "{ || " + AllTrim( tmp ) + " }" )
      jj ++
      Lput( aWks, Eval( bBlock ), lc( nRow + 2, jj ) )
   NEXT
   RETURN 1
*****************************************************************************
* End of Lputout() *
*****************************************************************************
* Purpose:     To determine the cell width for a given string
* Syntax:      Lgetwide( cString )
* Arguments:   cString      - Field or expression to output
* Notes:       Lgetwide() is used to extract a width portion from
*              a string passed to Lexport().  It is a STATIC
*              function only used in EXPORT.PRG and should not be
*              called from your application.
* Returns:     nWidth       - Number of characters wide
* Class:       Internal
**********************************************
STATIC FUNCTION lgetwide( cStr )
   LOCAL x1 := At( "|", cStr ), sarr := { }
   IF x1 == 0
      sarr := DBStruct()
      x1   := FieldPos( cStr )
      RETURN IF( x1 == 0, DEFAULT_WIDTH, sarr[ x1, DBS_LEN ] )
   ENDIF
   RETURN Val( SubStr( cStr, x1 + 1 ) )
*****************************************************************************
* Purpose:     To create an array from a semi-colon delimited string
* Syntax:      to_arr( cString )
* Arguments:   cString      - semi-colon delimited string
* Notes:       To_arr is a STATIC function used in Lexport() to
*              convert a semi-colon delimited string into an
*              array. The Summer '87 version requires a second
*              parameter, the array to fill.  This function is
*              used externally and should never be called by your
*              application.
* Returns:     aString      - Array of strings
* Class:       Internal
**********************************************
STATIC FUNCTION to_arr( cStr )
   LOCAL farr := { }, x
   WHILE ! Empty( cStr )
      x := At( ";", cStr )
      IF ! Empty( x )
         AAdd( farr, SubStr( cStr, 1, x - 1 ) )
         cStr := SubStr( cStr, x + 1 )
      ELSE
         AAdd( farr, cStr )
         cStr := ""
      ENDIF
   ENDDO
   RETURN farr
*****************************************************************************
* End of Lexport *
*****************************************************************************
